import java.io.PrintWriter;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.net.Socket;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * This class is the connection to the Websaver CGI script on the webserver.
 * It is used iternally by WebsaverOutputStream and WebsaverInputStream
 *
 * @author Brad Fitzpatrick &lt;bradfitz@bradfitz.com&gt;
 * @see WebsaverOutputStream
 * @see WebsaverInputStream
 *
 */
public class WebsaverConnection 
{
    private URL cgiHelper;
    private String directory;
    private boolean valid;

    private Socket sock;
    private PrintWriter writer;
    private BufferedReader in;


    /**
     *  Constructor for the WebsaverConnection
     *
     * @param url The full URL of the CGI script that will be doing the loading/saving.
     *            Note that if used from an applet, the hostname of the URL must be the 
     *            same as the one the applet was downloaded from.
     */
    public WebsaverConnection (String url)
    {
	this(url, null);
    }

    /**
     *  Constructor for the WebsaverConnection
     *
     * @param url The full URL of the CGI script that will be doing the loading/saving.
     *            Note that if used from an applet, the hostname of the URL must be the 
     *            same as the one the applet was downloaded from.
     *
     * @param directory  What directory on the server to store the files in.  This is
     *                   useful if you have different Java application/applets using
     *                   one Websaver.cgi and want to keep the files seperate, because
     *                   different application's files will invariably be of different types.
     */
    public WebsaverConnection (String url, String directory)
    {
        valid = false;
        try {
            cgiHelper = new URL(url);
            valid = true;
	    this.directory = directory;
        } catch (Exception e) { }
    }

    /**
     *  Method that sends a POST request to the webserver.
     *
     * @param reqHash  A Hashtable of String keys/values to send in a request
     *                 to the webserver.  The required keys include "mode", and
     *                 any other keys required for that mode.
     * 
     * @return A Hashtable representing the String keys/values of the server's 
     *         response.  The keys will be different depending on the mode.  The
     *         key "success" will always be one of "OK" or "FAIL".  If it's FAIL,
     *         the key "errmsg" will say what went wrong.
     */
    public Hashtable sendRequest (Hashtable reqHash)
    {
	if (directory != null)
	    reqHash.put("directory", directory);

        // buffers for the request and response
        StringBuffer req = new StringBuffer();
        StringBuffer res = new StringBuffer();

        // reset the response variable hash
        Hashtable resHash = new Hashtable();

        // build the request string
        for (Enumeration e = reqHash.keys(); e.hasMoreElements() ;) {
            String key = (String) e.nextElement();
            if (req.length() > 0)
                req.append('&');
            req.append(encodeURL(key));
            req.append('=');
            req.append(encodeURL((String)reqHash.get(key)));
        }

        try
        {
            int port = cgiHelper.getPort();
            if (port == -1) port = 80;
            sock = new Socket(cgiHelper.getHost(), port);
            writer = new PrintWriter(sock.getOutputStream());
            in  = new BufferedReader(new InputStreamReader(sock.getInputStream()));

            writer.print("POST " + cgiHelper.getFile() + " HTTP/1.0\r\n");
            writer.print("Host: "+ cgiHelper.getHost() + "\r\n");
            writer.print("Content-type: application/x-www-form-urlencoded\r\n");
            writer.print("Content-length: " + req.length() + "\r\n");
            writer.print("\r\n");
            writer.print(req);
            writer.flush();

            // read past the HTTP headers.  TODO: maybe examine them later for non-200 OK responses
            String line;
            boolean body = false;
            while (body==false && (line = in.readLine()) != null) {
                // see if we're past the HTTP headers
                if (line.length()==0) 
                    body=true;
            }

            // we're at the body.  let's scan it character at a time;
            StringBuffer key = new StringBuffer();
            StringBuffer value = new StringBuffer();
            boolean readValue = false;
            int i;
            while ((i = in.read()) != -1) {
                if (i == '=') {
                    readValue = true;
                } else if (i == '&') {
                    resHash.put(decodeURL(key.toString()), decodeURL(value.toString()));
                    key.setLength(0);
                    value.setLength(0);
                    readValue = false;
                } else if (readValue == true) {
                    value.append((char)i);
                } else if (readValue == false) {
                    key.append((char)i);
                }
            }
            // and put the last one in
            resHash.put(decodeURL(key.toString()), decodeURL(value.toString()));

            writer.close();
            in.close();
            sock.close();

        }
        catch (Exception e) { 
            System.out.println("connect exception: " + e.toString());
            valid = false;
        }

        return resHash;
    }

    /* given a 2-element char array of 2 hex digits, returns the decimal value */
    public static int hexValue (char [] c)
    {
        if (c.length != 2) return 0;

        int val = 0;   // initial return value
        for (int i=0; i<2; i++) {
            int thisval = 1;
            if (c[i] >= 'a' && c[i] <= 'f') { thisval = 10 + c[i] - 'a'; }
            else if (c[i] >= 'A' && c[i] <= 'F') { thisval = 10 + c[i] - 'A'; }
            else if (c[i] >= '0' && c[i] <= '9') { thisval = c[i] - '0'; }
            // increment the return value
            val += (i==0 ? 16 : 1) * thisval;
        }
        return val;
    }

    /* decodes a string from URL-escaped text into plain text */
    public static String decodeURL (String in)
    {
        StringBuffer ds = new StringBuffer();
        for (int i=0; i<in.length(); i++) {
            char c = in.charAt(i);
            if (c == '+') {
                // + chars get converted into spaces
                ds.append(' ');
            }  else if (c == '%') {
                // %hh gets converted into char w/ hex value 0xhh
                char [] hex = new char[2];
                hex[0] = in.charAt(i+1);
                hex[1] = in.charAt(i+2);
                int hv = hexValue(hex);
                ds.append((char)hv);
                i += 2;
            } else {
                // else, regular character
                ds.append(c);
            }
        }
        return ds.toString();
    }
        
    public static String decToHex (int i)
    {
	if (i > 255) return new String();
	StringBuffer sb = new StringBuffer();
	String s = "0123456789ABCDEF";
	sb.append(s.charAt(i/16));
	sb.append(s.charAt(i%16));
	return sb.toString();
    }
    
    /* encodes a string from plain text into URL-encoded text */
    public static String encodeURL (String in)
    {
	char tc;
        StringBuffer sb=new StringBuffer();
        for (int i=0; i<in.length(); i++)
	    {
		tc = in.charAt(i);
		if (tc=='_' || tc=='-' || tc=='.' || 
		    tc=='\\' || tc=='/' || tc==':' ||
		    (tc >= 'a' && tc <= 'z') ||
		    (tc >= 'A' && tc <= 'Z') ||
		    (tc >= '0' && tc <= '9')) sb.append(tc);
		else if (tc == ' ') sb.append('+');
		else sb.append("%" + decToHex((int)tc));
	    }
        return sb.toString();
    }
}
